# -*- coding: utf-8 -*-
"""
    gestikk, version 0.7

    gestikk provides mouse gestures for you.
    Copyright (C) 2007-2008 Friedrich Weber, Jonas Haag

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

BOOL_OPERATORS = ["and", "or", "not"]
COMPARISON_OPERATORS = ["is"]

OPERATORS = BOOL_OPERATORS + COMPARISON_OPERATORS

class ConditionError(Exception):
    pass

class Condor(object):
    """ a central class for ONE condition parsing action """
    def __init__(self, variables={}):
        self.variables = variables

    def run(self, string):
        """ run the string and return the return :-) """
        expressions = express(tokenize(string), self)
        # convert to a python expression
        pyexpr = self._as_pyexpr(expressions)
        if pyexpr == '()':
            return True
        # test
        if not test_pyexpr(pyexpr):
            raise ConditionError("Your condition may be malicious.")
        else:
            # eval it. I know that's dirty. But it should not be evil.
            return eval(pyexpr)

    def _as_pyexpr(self, expressions):
        if isinstance(expressions, list):
            return '('+' '.join([self._as_pyexpr(expr) for expr in expressions])+')'
        else:
            return str(expressions.as_bool())
        
    def get_variable(self, name):
        return self.variables[name] # TODO?

class Expression(object):
    def __init__(self, condor, comparison_op):
        self.operator = comparison_op
        self.content = []
        self.condor = condor
        self.end = False

    def append(self, stuff):
        self.content.append(stuff)
        if len(self.content) == 2:
            self.end = True
    
    def get(self, n):
        """ get the n-th content token in the right type. evaluate variables. """
        c = self.content[n]
        if isinstance(c, VariableToken):
            return self.condor.get_variable(c)
        else:
            return c

    def as_bool(self):
        """ return the return of the expression as bool """
        get = self.get
        if self.operator == 'is': # content[0] == content[1]?
            return get(0) == get(1)
        elif self.operator == 'startswith':
            return get(0).startswith(get(1))
        elif self.operator == 'endswith':
            return get(0).endswith(get(1))
        elif self.operator == 'contains':
            return get(1) in get(0)

    def __repr__(self):
        return '<Expression Operator "%s", content is %s>' % (self.operator, str(self.content))

class BoolExpression(object):
    def __init__(self, name):
        self.name = name
        self.end = True

    def __repr__(self):
        return '<Boolean Expression "%s">' % self.name

    def as_bool(self):
        return self.name

def typisize(token):
    if token.isdigit():
        return int(token)
    elif token[-1] == '"' and token[0] == '"':
        return token[1:-1]
    else:
        return VariableToken(token)

def test_pyexpr(string):
    """ test a pyexpr whether it does not use disallowed tokens
        return True if everything's ok
    """
    TOKENS = ["True", "False", " ", "(", ")", "and", "or", "not"]
    pointer = 0
    while pointer < len(string):
        for t in TOKENS:
            if string[pointer:pointer+len(t)] == t:
                pointer += len(t)
                break
        else:
            raise Exception("disallowed token in '%s'" % string)
            return False
    return True

def tokenize(string):
    """ tokenize the string to a list of strings. """
    stack = [[]]
    token = ''
    pointer = 0
    quoted = False
    while pointer != len(string):
        curr = string[pointer]
        if not quoted and string[pointer-1] != '\\':
            if curr == '(':
                stack.append([])
                token = ''
            elif curr == ')':
                token = typisize(token)
                stack[-1].append(token)
                token = ''
                last = stack.pop(-1)
                stack[-1].append(last)
            elif curr == ' ':
                token = typisize(token)
                stack[-1].append(token)
                token = ''
            elif curr == '"':
                quoted = not quoted
                token += '"'
            else:
                token += curr
        else:
            if curr == '"':
                quoted = not quoted
            token += curr
        pointer += 1
    if token:
        stack[-1].append(typisize(token))
    return stack[-1]

def express(tokens, condor):
    """ convert a list of tokens (produced by `tokenize`) to a list of Expressions. """
    expr = None
    expressions = []
    stack = []
    for token in tokens:
        if expr is None:
            if token in COMPARISON_OPERATORS:
                assert expr is None
                expr = Expression(condor, token)
                for st in stack:
                    expr.append(st)
                stack = []
            elif token in BOOL_OPERATORS:
                expr = BoolExpression(token)
            elif isinstance(token, list):
                expressions.append(express(token, condor))
            else:
                stack.append(token)
        elif not expr.end:
            expr.append(token)
        if expr and expr.end:
            expressions.append(expr)
            expr = None
    return expressions

class VariableToken(str):
    pass

if __name__ == '__main__':
    c = Condor({'activeApplication':'Blubb', 'activeWorkspace':1})
    print c.run('activeApplication is "Blubb" and (activeWorkspace is 1 or activeWorkspace is 3)')
